from pathlib import Path
from shutil import copyfile

import multiprocessing as mp

import numpy as np
import wfdb
import yaml

import lifedataphysionetchallenge as lpc

DATA_PREPROCESSED_PATH = lpc.get_project_root() / "data/preprocess"
DATA_DOWER_PATH = lpc.get_project_root() / "data/dower_transformation"
DATA_DOWER_PATH.mkdir(exist_ok=True, parents=True)

# Get params from params.yml
with open(lpc.get_project_root() / "params.yaml") as par_file:
    params = yaml.safe_load(par_file)
WFDB_FORMAT = params["DOWER"]["WFDB_FORMAT"]
DOWER_TRANSFORM = params["DOWER"]["DOWER_TRANSFORM"]


def dowertransform(data, header):
    # Dower transformation only needs leads ['I', 'II', 'V1', 'V2', 'V3', 'V4', 'V5', 'V6']
    # They're at indexes [0, 1, 6, 7, 8, 9, 10, 11]
    # Remaining leads:
    remain_leads = np.array([0, 1, 6, 7, 8, 9, 10, 11])

    ecgDower = data[:, remain_leads]

    # NOTE: Link to Matlab transformation:
    # https://github.com/RuneKR/htm-ecg/blob/ad3d525497b8ebaf2fb95a86fb1b9241a9cfc263/matlab/lib/functions/ECG2VCGFcn.m

    # NOTE: ECG = Matlab struct or something
    # NOTE: Take fields and insert them to matrix + transpose them
    # NOTE:

    # Inverse Dower matrix: VCG = Minv * ECG
    #                       [X Y Z] = Minv * [I II V1 V2 V3 V4 V5 V6]'
    #          I      II     V1     V2     V3     V4     V5     V6

    invDower = np.asarray([
        [0.156, -0.010, -0.172, -0.074, 0.122, 0.231, 0.239, 0.194],  # X
        [-0.227, 0.887, 0.057, -0.019, -0.106, -0.022, 0.041, 0.048],  # Y
        [0.022, 0.102, -0.229, -0.310, -0.246, -0.063, 0.055, 0.108]  # Z
    ])

    # NOTE: According to Franck the VCG coordinates are:
    # X: left lateral direction
    # Y: inferior direction
    # Z: posterior direction

    # According to the Dower matrix, the lower row represents the anterior
    # direction, hence -Z relative to Franck.
    # Transform the (inverse) Dower matrix to Franck's lead system, which is
    # right-handed instead of the left-handed Dower system:

    VCG = np.dot(invDower, ecgDower.T)
    VCG = VCG.T

    # plt = hv.Scatter3D((VCG.T[0].flat, VCG.T[1].flat, VCG.T[2].flat)).opts(
    #     opts.Scatter3D(azimuth=40, elevation=20, color='z', s=50, cmap='fire'))

    # hv.save(plt, f'{lpc.get_project_root() / "data" / "test" / "dower"}.html')

    return(VCG, header)


def wfdb_save(data, filename):
    wfdb.wrsamp(filename, fs=data.fs, units=data.units, sig_name=data.sig_name,
                p_signal=data.p_signal, fmt=(len(data.sig_name) * [WFDB_FORMAT]), base_time=data.base_time,
                base_date=data.base_date, write_dir=DATA_DOWER_PATH, comments=data.comments)


def dower_worker(file):
    # Load data from drive
    data = wfdb.rdrecord(str(file))
    ecg_data = data.p_signal.T
    # TODO If any changes to the header data is necessary, read the header here
    header = ""

    if DOWER_TRANSFORM == True:
        # Transform data
        VCG, header = dowertransform(ecg_data.T, header)
        # change lead names
        data.sig_name = ['X', 'Y', 'Z']
    elif DOWER_TRANSFORM == False:
        VCG = data.p_signal

    # NOTE: wfdb.rdrecord() reads data and header files
    # Store transformed data
    # Units can't be empty for wfdb format, to target this problem unit '' was set
    data.units = VCG.shape[1] * ['']
    data.p_signal = VCG
    wfdb_save(data, file.stem)


if __name__ == "__main__":
    files = [i.with_suffix('') for i in DATA_PREPROCESSED_PATH.rglob('*.hea')]

    with mp.Pool() as pool:
        pool.map(dower_worker, files)
