# %% Code for setting random seed must run first
import csv
from datetime import datetime
from pathlib import Path

import numpy as np
import pandas as pd
import tensorflow as tf
import yaml
from numpy.random import seed
from sklearn.preprocessing import MultiLabelBinarizer
from sklearn.utils.class_weight import compute_class_weight
from tensorflow import keras
from tensorflow.keras.layers import (Activation, Add, BatchNormalization,
                                     Conv1D, Dense, Dropout, Flatten, Input,
                                     LeakyReLU, MaxPooling1D)
from tensorflow.keras.models import Model, Sequential
from tensorflow.keras.optimizers import SGD, Adam
from tensorflow.keras.wrappers.scikit_learn import KerasClassifier

import lifedataphysionetchallenge as lpc
import lifedataphysionetchallenge.train_test_split
import lifedataphysionetchallenge.util.gpu_utils
from lifedataphysionetchallenge.util.ecg_data_generator import ECGDataGenerator


def model_training(input_directory, output_directory):
    with open(lpc.get_project_root() / "params.yaml") as file:
        params = yaml.safe_load(file)

    # TODO configure random state to be actually used
    RANDOM_STATE = params["RANDOM_STATE"]

    # Seed numpy
    seed(RANDOM_STATE)
    # Seed tensorflow
    tf.random.set_seed(RANDOM_STATE)

    # %%
    from tensorflow.keras.layers import InputLayer, Input, Conv1D, Add, Dense, Flatten, Dropout, Activation, LeakyReLU, MaxPooling1D, BatchNormalization

    lifedataphysionetchallenge.util.gpu_utils.mask_unused_gpus(
        count_gpus_unmasked=1, minimum_available_ram=9000)

    # %%
    print(f"Tensorflow Version: ", tf.__version__)

    # Inputs
    DATA_DOWER_TRANSFORMED = input_directory
    WEIGHTS_CSV = lpc.get_project_root() / "data-analysis" / \
        "metadata-analysis" / "class_weights.csv"

    # Outputs
    DATA_MODEL = output_directory
    DATA_MODEL.mkdir(parents=True, exist_ok=True)

    # %%
    BATCH_SIZE = params["MODEL_TRAINING"]["BATCH_SIZE"]
    EPOCHS = params["MODEL_TRAINING"]["EPOCHS"]
    LEADS = params["MODEL_TRAINING"]["LEADS"]
    LEARNING_RATE = params["MODEL_TRAINING"]["LEARNING_RATE"]
    MODEL_ARCH = params["MODEL_TRAINING"]["MODEL_ARCH"]
    DURATION = params["PREPROCESS"]["DURATION"]
    SAMPLERATE = params["PREPROCESS"]["SAMPLERATE"]
    # %%

    #######
    # load training data
    ######

    # The list of all record IDs:
    record_paths = list(DATA_DOWER_TRANSFORMED.glob("*.hea"))

    # %%

    train_ids = lifedataphysionetchallenge.train_test_split.get_train_ids()
    test_ids = lifedataphysionetchallenge.train_test_split.get_test_ids()

    # %%
    # Sort the record paths into train paths and test paths by
    # checking if the stem of the path is part of the train or test ID list
    train_paths = [train_path
                   for train_path in record_paths
                   if train_path.stem in train_ids
                   ]

    validation_paths = [val_path
                        for val_path in record_paths
                        if val_path.stem in test_ids
                        ]
    # %%
    record_length = DURATION * SAMPLERATE  # Length of a record

    # %%
    train_data_gen = ECGDataGenerator(train_paths, dim=(record_length, LEADS))
    val_data_gen = ECGDataGenerator(
        validation_paths, dim=(record_length, LEADS))
    # %%

    # Create dict with weight in unbalanced and randowm form for tests
    # class_weigth = {key: value for key, value in enumerate(np.random.rand(1, train_data_gen.n_classes)[0])}
    # Create class weight dictonary from weights.csv. That file was created in metadata_analysis
    unsorted_weights = pd.read_csv(WEIGHTS_CSV, index_col=0)
    sorted_weights = unsorted_weights.reindex(
        [int(idx) for idx in train_data_gen.multi_label_binarizer.classes_])
    class_weights = {key: value[0]
                     for key, value in enumerate(sorted_weights.values)}

    # Configuring Tensorboard
    tensorboard_log_dir = DATA_MODEL / "tensorboard_logs" / \
        datetime.now().strftime("%Y%m%d-%H%M%S")
    file_writer = tf.summary.create_file_writer(str(tensorboard_log_dir))
    file_writer.set_as_default()

    tensorboard_callback = keras.callbacks.TensorBoard(
        log_dir=tensorboard_log_dir, histogram_freq=1)

    # %%

    class ResidualUnit(object):
        """Residual unit block (unidimensional).
        Parameters
        ----------
        n_samples_out: int
            Number of output samples.
        n_filters_out: int
            Number of output filters.
        kernel_initializer: str, otional
            Initializer for the weights matrices. See Keras initializers. By default it uses
            'he_normal'.
        dropout_rate: float [0, 1), optional
            Dropout rate used in all Dropout layers. Default is 0.8
        kernel_size: int, optional
            Kernel size for convolutional layers. Default is 17.
        preactivation: bool, optional
            When preactivation is true use full preactivation architecture proposed
            in [1]. Otherwise, use architecture proposed in the original ResNet
            paper [2]. By default it is true.
        postactivation_bn: bool, optional
            Defines if you use batch normalization before or after the activation layer (there
            seems to be some advantages in some cases:
            https://github.com/ducha-aiki/caffenet-benchmark/blob/master/batchnorm.md).
            If true, the batch normalization is used before the activation
            function, otherwise the activation comes first, as it is usually done.
            By default it is false.
        activation_function: string, optional
            Keras activation function to be used. By default 'relu'.
        References
        ----------
        .. [1] K. He, X. Zhang, S. Ren, and J. Sun, "Identity Mappings in Deep Residual Networks,"
            arXiv:1603.05027 [cs], Mar. 2016. https://arxiv.org/pdf/1603.05027.pdf.
        .. [2] K. He, X. Zhang, S. Ren, and J. Sun, "Deep Residual Learning for Image Recognition," in 2016 IEEE Conference
            on Computer Vision and Pattern Recognition (CVPR), 2016, pp. 770-778. https://arxiv.org/pdf/1512.03385.pdf
        """

        def __init__(self, n_samples_out, n_filters_out, kernel_initializer='he_normal',
                     dropout_rate=0.8, kernel_size=17, preactivation=True,
                     postactivation_bn=False, activation_function='relu'):
            self.n_samples_out = n_samples_out
            self.n_filters_out = n_filters_out
            self.kernel_initializer = kernel_initializer
            self.dropout_rate = dropout_rate
            self.kernel_size = kernel_size
            self.preactivation = preactivation
            self.postactivation_bn = postactivation_bn
            self.activation_function = activation_function

        def _skip_connection(self, y, downsample, n_filters_in):
            """Implement skip connection."""
            # Deal with downsampling
            if downsample > 1:
                y = MaxPooling1D(
                    downsample, strides=downsample, padding='same')(y)
            elif downsample == 1:
                y = y
            else:
                raise ValueError("Number of samples should always decrease.")
            # Deal with n_filters dimension increase
            if n_filters_in != self.n_filters_out:
                # This is one of the two alternatives presented in ResNet paper
                # Other option is to just fill the matrix with zeros.
                y = Conv1D(self.n_filters_out, 1, padding='same',
                           use_bias=False, kernel_initializer=self.kernel_initializer)(y)
            return y

        def _batch_norm_plus_activation(self, x):
            if self.postactivation_bn:
                x = Activation(self.activation_function)(x)
                x = BatchNormalization(center=False, scale=False)(x)
            else:
                x = BatchNormalization()(x)
                x = Activation(self.activation_function)(x)
            return x

        def __call__(self, inputs):
            """Residual unit."""
            x, y = inputs
            n_samples_in = y.shape[1]
            downsample = n_samples_in // self.n_samples_out
            n_filters_in = y.shape[2]
            y = self._skip_connection(y, downsample, n_filters_in)
            # 1st layer
            x = Conv1D(self.n_filters_out, self.kernel_size, padding='same',
                       use_bias=False, kernel_initializer=self.kernel_initializer)(x)
            x = self._batch_norm_plus_activation(x)
            if self.dropout_rate > 0:
                x = Dropout(self.dropout_rate)(x)

            # 2nd layer
            x = Conv1D(self.n_filters_out, self.kernel_size, strides=downsample,
                       padding='same', use_bias=False,
                       kernel_initializer=self.kernel_initializer)(x)
            if self.preactivation:
                x = Add()([x, y])  # Sum skip connection and main connection
                y = x
                x = self._batch_norm_plus_activation(x)
                if self.dropout_rate > 0:
                    x = Dropout(self.dropout_rate)(x)
            else:
                x = BatchNormalization()(x)
                x = Add()([x, y])  # Sum skip connection and main connection
                x = Activation(self.activation_function)(x)
                if self.dropout_rate > 0:
                    x = Dropout(self.dropout_rate)(x)
                y = x
            return [x, y]

    signal = Input(shape=(record_length, LEADS),
                   dtype=np.float32, name='signal')

    kernel_size = 8
    kernel_initializer = 'he_normal'

    x = signal
    x = Conv1D(64, kernel_size, padding='same', use_bias=False,
               kernel_initializer=kernel_initializer)(x)
    x = BatchNormalization()(x)
    x = Activation('relu')(x)
    x, y = ResidualUnit(1024, 128, kernel_size=kernel_size,
                        kernel_initializer=kernel_initializer)([x, x])
    x, y = ResidualUnit(256, 196, kernel_size=kernel_size,
                        kernel_initializer=kernel_initializer)([x, y])
    x, y = ResidualUnit(64, 256, kernel_size=kernel_size,
                        kernel_initializer=kernel_initializer)([x, y])
    x, _ = ResidualUnit(16, 320, kernel_size=kernel_size,
                        kernel_initializer=kernel_initializer)([x, y])
    x = Flatten()(x)

    diagn = Dense(train_data_gen.n_classes, activation='sigmoid',
                  kernel_initializer=kernel_initializer)(x)
    #model = Model(signal, diagn)
    model = Model(inputs=signal, outputs=diagn)

    # %%
    # model compilation

    adam = Adam(learning_rate=0.001, beta_1=0.9, beta_2=0.999, amsgrad=False)

    model.compile(loss='categorical_crossentropy',
                  optimizer=adam,
                  metrics=["categorical_accuracy"])

    model.summary()

    # %%
    # model training
    model_history = model.fit(train_data_gen, validation_data=val_data_gen,
                              epochs=EPOCHS, verbose=2, callbacks=[tensorboard_callback], class_weight=class_weights)

    # save model
    model.save(str(DATA_MODEL/f'model.h5'))
    print('model saved')

    # save tensorboard-history
    pd.DataFrame(model_history.history).to_csv(str(DATA_MODEL)+"/history.csv")

    """
    # NOTE: What is the difference of model.save(str(DATA_MODEL/f'model.h5')) to
    model_json = model.to_json()
    json_file = open(self.ModelName, "w")
    json_file.write(model_json)
    model.save_weights(self.WeightName)
    """

    # %%


if __name__ == "__main__":
    # WindowsPath('c:/users/peter.kienle/physionet2020/data/dower_transformation')
    # WindowsPath('c:/users/peter.kienle/physionet2020/data/model_training')

    # Create input and output directories which will be used as arguments for the
    # model_training function
    input_directory = lpc.get_project_root() / "data/dower_transformation"
    output_directory = lpc.get_project_root() / "data/model_training"

    model_training(input_directory, output_directory)
