# %%
# Create predictions in the desired physionet format, see section "Algorithms" here: https://physionetchallenges.github.io/2020/
# %%
import yaml
import lifedataphysionetchallenge as lpc
import lifedataphysionetchallenge.model_training
import lifedataphysionetchallenge.train_test_split
from lifedataphysionetchallenge.util.ecg_data_generator import ECGDataGenerator
from lifedataphysionetchallenge.util import snomed
import csv
import logging
import tensorflow as tf
import shutil
from sklearn.metrics import multilabel_confusion_matrix
import numpy as np

# %%

# TODO Use log levels properly
# Setting the level this way should work, but it doesn't, therefore I made all messages warnings
logging.basicConfig(level=logging.INFO)
logging.info("Starting Prediction of TestData")
logging.warning("Starting Prediction of TestData")

# %%
# Inputs
DATA_MODEL = lpc.get_project_root() / "data/model_training/model.h5"
DATA_DOWER_TRANSFORMED = lpc.get_project_root() / "data/dower_transformation"
# Needed for the original headers, see https://git.rz.uni-augsburg.de/swtpvs-lifedata/physionet2020/issues/38#note_29977
DATA_RAW = lpc.get_project_root() / "data/raw"
# FIXME make this something that's created during model training
# It's the threshold above which we classify this disease as present
PREDICTION_THRESHOLD = 0.7

# Outputs
DATA_TEST_PREDICTIONS = lpc.get_project_root() / "data/predicted_testdata/"
DATA_TEST_PREDICTIONS_Y_PRED = DATA_TEST_PREDICTIONS / "y_pred"
DATA_TEST_PREDICTIONS_Y_PRED.mkdir(parents=True, exist_ok=True)
DATA_TEST_PREDICTIONS_Y_TRUE = DATA_TEST_PREDICTIONS / "y_true"
DATA_TEST_PREDICTIONS_Y_TRUE.mkdir(parents=True, exist_ok=True)
# %%
######
# Load params
######

with open(lpc.get_project_root() / "params.yaml") as file:
    params = yaml.safe_load(file)

DURATION = params["PREPROCESS"]["DURATION"]
SAMPLERATE = params["PREPROCESS"]["SAMPLERATE"]
LEADS = params["MODEL_TRAINING"]["LEADS"]

record_length = DURATION * SAMPLERATE

# %%
#####
# Load the model
#####
model = lpc.model_training.load_model(DATA_MODEL)

# %%
######
# Load the list of paths of data to be predicted
######

# The list of all record IDs:
record_paths = list(DATA_DOWER_TRANSFORMED.glob("*.hea"))

# %%

logging.warning("Collecting Test Record File Paths")

test_ids = lifedataphysionetchallenge.train_test_split.get_test_ids()
validation_paths = [val_path
                    for val_path in record_paths
                    if val_path.stem in test_ids
                    ]

# %%
# Don't shuffle, otherwise we can't recall which ID belongs to this prediction
# Also make batch_size=1, otherwise the last couple of paths which don't fit in a complete
# batch will be ignored
ecgDataGenerator = ECGDataGenerator(validation_paths, dim=(record_length, LEADS),
                                    shuffle=False, batch_size=1)

# %%
logging.warning("Predicting test data")
# Create predictions, this takes a while
preds = np.round(model.predict(ecgDataGenerator), decimals=3)
# preds

# %%
y_true = np.concatenate([batch[1] for batch in ecgDataGenerator], axis=0)

# %%


def write_prediction_output_csv(csv_directory_path, record_id, prediction, prediction_threshold):
    """
    # TODO Document expected inputs
    Format the predictions in the following format
    See "Algorithms" here https://physionetchallenges.github.io/2020/
    #Record ID
    SNOMED Code, SNOMED Code, SNOMED Code,
    1, 0, 0, 1,
    0.9, 0.6, 0.2, 
    """
    with (csv_directory_path / (str(record_id) + ".csv")).open("w") as csv_file:
        csvWriter = csv.writer(csv_file)
        csvWriter.writerow([f"#{record_id}"])
        csvWriter.writerow(ecgDataGenerator.multi_label_binarizer.classes_)
        csvWriter.writerow([1 if class_confidence > prediction_threshold else 0
                            for class_confidence in prediction])
        csvWriter.writerow(prediction)
# %%


# NOTE Matching the IDs with their predictions here after the fact is a bit sketchy, could lead to mismatches
# Maybe try loading all predictions at once or something (check if that needs too much RAM)
logging.warning(
    f"Writing prediction to files in {str(DATA_TEST_PREDICTIONS_Y_PRED)}")
for i, test_id in enumerate(test_ids):
    # Save prediction

    write_prediction_output_csv(
        DATA_TEST_PREDICTIONS_Y_PRED, test_id, preds[i], PREDICTION_THRESHOLD)

logging.warning(
    f"Copying raw headers into {str(DATA_TEST_PREDICTIONS_Y_TRUE)}")
raw_headers = list(DATA_RAW.glob("*.hea"))
# Copy raw data headers, so that during the evaluation stage it can be evaluated with the official physionet methods
for raw_header in raw_headers:
    if raw_header.stem in test_ids:
        shutil.copy(raw_header, DATA_TEST_PREDICTIONS_Y_TRUE/raw_header.name)


# %%
