# %%
import logging
import multiprocessing as mp
from pathlib import Path

import numpy as np
import wfdb
from scipy.io import loadmat
from scipy import signal
import yaml
import lifedataphysionetchallenge.train_test_split

from sklearn.preprocessing import normalize
from sklearn.preprocessing import MinMaxScaler

# Get signalpaths
import lifedataphysionetchallenge as lpc
DATA_RAW_PATH = lpc.get_project_root() / "data" / "raw"
DATA_PREPROCESSED_PATH = lpc.get_project_root() / "data" / "preprocess"
DATA_PREPROCESSED_PATH.mkdir(exist_ok=True, parents=True)

# Get params from params.yml
with open(lpc.get_project_root() / "params.yaml") as par_file:
    params = yaml.safe_load(par_file)
SAMPLERATE = params["PREPROCESS"]["SAMPLERATE"]
DURATION = params["PREPROCESS"]["DURATION"]
WFDB_FORMAT = params["PREPROCESS"]["WFDB_FORMAT"]
SCALING = params["PREPROCESS"]["SCALING"]
SET_SIZE_FOR_DEBUGGING = params["SET_SIZE_FOR_DEBUGGING"]
ONLY_SNOMED_CLASSES = set(params["ONLY_SNOMED_CLASSES"])


def preprocess(data, header):
    logging.debug(f"Resample data to {SAMPLERATE}")
    # sample rate in input signal is stored in header[0].split()[2]
    resampled_data = resampling(data, int(header[0].split()[2]))

    logging.debug(f"Norm signals with {SCALING}")
    # scaled_data = uniting(resampled_data)
    scaled_data = norm_signals(resampled_data, SCALING)

    logging.debug(f'Segmenting to {DURATION} seconds signals')
    segmented_data = segmenting(scaled_data)
    # Storing the data via wfdb.wrsamp later needs this data to be transposed

    return np.array(segmented_data).T, header


def resampling(data, sr):
    """
    """
    # Note: There are 3 scypy methods for downsampling
    # - Methode 1: scipy.signal.resample (FFT) -> used method
    #    - https://docs.scipy.org/doc/scipy/reference/generated/scipy.signal.resample.html#scipy.signal.resample
    # - Methode 2: scipy.signal.decimate (FIR or IIR filter)
    # - Methode 3: scipy.resample_poly (polyphase filtering and FIR Filter)

    # Calculate downsample factor
    sample_factor = sr / SAMPLERATE
    # calculate number of signals after downsampling
    res_sig_nr = [int(len(lead) / sample_factor) for lead in data]

    # plot(data, sr, [signal.resample(lead, sig_nr) for lead, sig_nr in zip(data, res_sig_nr)],
    #      "resampling difference", writedir=lpc.get_project_root() / "data" / "test")

    # resampling
    return [signal.resample(lead, sig_nr) for lead, sig_nr in zip(data, res_sig_nr)]


def segmenting(signal):
    """
    """
    nr_sample = int(SAMPLERATE * DURATION)
    # TODO: timeshift is for segementing with tumbling window
    # Use this later
    timeshift = 0
    firstsample = int(timeshift * SAMPLERATE)
    lastsample = int(firstsample + nr_sample)

    segment = [lead[firstsample:lastsample] for lead in signal]
    missing_values = [nr_sample - len(lead) for lead in segment]

    return [np.pad(lead, (0, missing_values_lead), 'constant', constant_values=(0, 0)) for lead, missing_values_lead in zip(segment, missing_values)]


def norm_signals(data, conf=None):
    """
    """
    # Link to package https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.MinMaxScaler.html
    if conf == "norm":
        # create normaliser
        scale = MinMaxScaler(feature_range=(0, 1))
    elif conf == "scale":
        # create scaler
        scale = MinMaxScaler(feature_range=(-1, 1))
    elif conf == "uniting":
        unitscaler = 1000
        return [lead / unitscaler for lead in data]
    else:
        return data

    # apply transform
    scaled = [scale.fit_transform(lead.reshape(-1, 1)) for lead in data]
    scaled = [lead.reshape(-1, ) for lead in scaled]
    return scaled


def preprocess_worker(file):
    # Load matrix data
    logging.debug(f'Read record values from {file}')
    matrix = loadmat(file.with_suffix('.mat'))
    data = np.asarray(matrix['val'], dtype=np.float64)

    # Load header as string
    logging.debug(f'Read heder from file {file}')
    with open(file.with_suffix('.hea'), 'r') as f:
        header = f.readlines()

    ######
    # Parse Header string and adapt it to our preprocessing changes
    #####
    # Signalinfo
    # We're changing the samplerate during preprocessing
    samplerate = SAMPLERATE
    base_time = header[0].split()[-1]
    base_date = header[0].split()[-2]

    # Leadinfo
    leadinfo = header[1:-6]
    lead_names = [lead.split()[-1] for lead in leadinfo]
    units = [lead.split()[2].split("/")[1] for lead in leadinfo]
    # Change units to nothing cause they have none after normalization
    # During preprocess signal was scaled and is now unitless
    unit = ['']
    logging.debug(f"Changed unit to {unit}")
    units = len(lead_names) * ['']

    # Filter comments from the header string. Comments start with '#'
    comments = []
    for line in header:
        if line.startswith("#"):
            comments = comments + [line]

    #####
    # The Actual Preprocessing
    #####
    logging.debug("Start preprocessing")
    segmented_data, header = preprocess(data, header)

    # Store the changed signal and headers
    # NOTE: we're droppping basetime unintentionally to save problems during storing
    logging.debug("Save wfdb file")
    wfdb.wrsamp(file.stem, fs=samplerate, units=units, sig_name=lead_names,
                p_signal=segmented_data, fmt=(
                    len(lead_names) * [WFDB_FORMAT]),
                write_dir=str(DATA_PREPROCESSED_PATH), comments=comments)


# %%

if __name__ == "__main__":
    logging.basicConfig(level=logging.INFO, filemode='w', datefmt='%d-%b-%y %H:%M:%S',
                        format='%(asctime)s - %(name)s - %(levelname)s - %(message)s')

    logging.debug(f'Get files in {DATA_RAW_PATH}')
    files = [filename.with_suffix('')
             for filename in DATA_RAW_PATH.rglob('*.hea')]

    # Check if we're debugging or restricting SNOMED CT IDs
    if (SET_SIZE_FOR_DEBUGGING != -1) | (ONLY_SNOMED_CLASSES != set([])):
        logging.debug("Load debug dataset")
        # If so, only preprocess data which is in the train/test ids
        debugging_files = []
        train_test_ids = lifedataphysionetchallenge.train_test_split.get_all_ids()
        # Filter only the files part of the train_test_ids
        debugging_files = [file
                           for file in files
                           if file.stem in train_test_ids]
        files = debugging_files

    logging.debug("preprocess files")

    with mp.Pool() as pool:
        pool.map(preprocess_worker, files)


# %%
