import numpy as np
import tensorflow.keras as keras
import wfdb
from lifedataphysionetchallenge.util import snomed
from lifedataphysionetchallenge.util.ecg_multi_label_binarizer import get_binarizer

class ECGDataGenerator(keras.utils.Sequence):
    'Generates data for Keras'
    def __init__(self, list_matrix_paths, batch_size=32, dim=(32,32,32),
                shuffle=True):
        'Initialization'
        self.dim = dim
        self.batch_size = batch_size
        self.list_matrix_paths = list_matrix_paths
        self.multi_label_binarizer = get_binarizer()
        self.n_classes = len(self.multi_label_binarizer.classes_)
        self.shuffle = shuffle
        self.on_epoch_end()

    def __len__(self):
        'Denotes the number of batches per epoch'
        return int(np.floor(len(self.list_matrix_paths) / self.batch_size))

    def __getitem__(self, index):
        'Generate one batch of data'
        # Generate indexes of the batch
        indexes = self.indexes[index*self.batch_size:(index+1)*self.batch_size]

        # Find list of IDs
        list_matrix_paths_in_batch = [self.list_matrix_paths[k] for k in indexes]

        # Generate data
        X, y = self.__data_generation(list_matrix_paths_in_batch)

        return X, y

    def on_epoch_end(self):
        'Updates indexes after each epoch'
        self.indexes = np.arange(len(self.list_matrix_paths))
        if self.shuffle == True:
            np.random.shuffle(self.indexes)

    def __data_generation(self, list_matrix_paths_in_batch):
        'Generates data containing batch_size samples' # X : (n_samples, *dim)
        # Initialization
        X = np.empty((self.batch_size, *self.dim))
        y = np.empty((self.batch_size, self.n_classes), dtype=int)

        # Generate data
        for i, matrix_path in enumerate(list_matrix_paths_in_batch):
            ######
            # Load sample
            ######

            # wfdb.rdrecord expects a string pointing to a path, however without a file ending
            # The file ending is stripped by setting the suffix to ""
            ecg_record = wfdb.rdrecord(str(matrix_path.with_suffix("")))
            # TODO Check the [0] here, it shouldn't be necessary, but without it this error occurs: https://git.rz.uni-augsburg.de/swtpvs-lifedata/physionet2020/-/jobs/23429
            #X[i] = ecg_record.p_signal[:, 1][0]
            # Use syntax below when multiple leads of the record are to be used again
            X[i,] = ecg_record.p_signal

            # Load class
            header = wfdb.rdheader(str(matrix_path.with_suffix("")))
            dx_ids = snomed.get_snomed_ids_from_header_comments(header.comments)
            # Transform the list of IDs into a vector of 1s and 0s, see https://scikit-learn.org/stable/modules/generated/sklearn.preprocessing.MultiLabelBinarizer.html
            # The result of `transform` is a list with a single element, take that element out of the list via `[0]`
            y[i] = self.multi_label_binarizer.transform([dx_ids])[0]

        return X, y