import subprocess as sp
import os
import tensorflow as tf

def mask_unused_gpus(count_gpus_unmasked, minimum_available_ram = 1024):
    """
    Autmatically hide GPUs in use from Tensorflow / Keras via setting
    the env variable `CUDA_VISIBLE_DEVICES` in the operating system
    
    Parameters:
    -----------
    
    count_gpus_unmasked : int
      Count of GPUs to leave unmasked
    minimum_available_ram : int
      Minimum RAM in MB
    """
    try:
        CUDA_VISIBLE_DEVICES = os.environ["CUDA_VISIBLE_DEVICES"]
        print(f'Using already set devices {CUDA_VISIBLE_DEVICES}')
        return
    except KeyError:
        pass
        
    
    ACCEPTABLE_AVAILABLE_MEMORY = minimum_available_ram
    COMMAND = "nvidia-smi --query-gpu=memory.free --format=csv"

    try:
        _output_to_list = lambda x: x.decode('ascii').split('\n')[:-1]
        memory_free_info = _output_to_list(sp.check_output(COMMAND.split()))[1:]
        memory_free_values = [int(x.split()[0]) for i, x in enumerate(memory_free_info)]
        available_gpus = [i for i, x in enumerate(memory_free_values) if x > ACCEPTABLE_AVAILABLE_MEMORY]

        if len(available_gpus) < count_gpus_unmasked: raise ValueError('Found only %d usable GPUs in the system' % len(available_gpus))
        os.environ["CUDA_VISIBLE_DEVICES"] = ','.join(map(str, available_gpus[:count_gpus_unmasked]))
    except Exception as e:
        print('"nvidia-smi" is probably not installed. GPUs are not masked', e)

################################
# Only grab necessary GPU RAM
# This is executed as soon as this package is imported
################################
gpus = tf.config.experimental.list_physical_devices('GPU')
if gpus:
    try:
        # Currently, memory growth needs to be the same across GPUs
        for gpu in gpus:
            tf.config.experimental.set_memory_growth(gpu, True)
        logical_gpus = tf.config.experimental.list_logical_devices('GPU')
        print(f"Set `Memory Growth` for {len(gpus)} Physical GPUs, {len(logical_gpus)} Logical GPUs")
    except RuntimeError as e:
        # Memory growth must be set before GPUs have been initialized
        print(e)